@@ -15,11 +15,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with NILFS; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
- * Written by Koji Sato <koji@osrg.net>.
+ * Written by Koji Sato.
  *
  * Maintained by Ryusuke Konishi <konishi.ryusuke@lab.ntt.co.jp> from 2008.
  */
@@ -56,6 +52,10 @@
 #include <sys/stat.h>
 #endif	/* HAVE_SYS_STAT_H */
 
+#if HAVE_TIME_H
+#include <time.h>	/* timespec */
+#endif	/* HAVE_TIME_H */
+
 #if HAVE_SYSLOG_H
 #include <syslog.h>
 #endif	/* HAVE_SYSLOG_H */
@@ -63,6 +63,7 @@
 #include <errno.h>
 #include <assert.h>
 #include "nilfs.h"
+#include "util.h"
 #include "cldconfig.h"
 
 
@@ -138,8 +139,8 @@ static int nilfs_cldconfig_get_ulong_argument(char **tokens, size_t ntoks,
 	return 0;
 }
 
-static int nilfs_cldconfig_get_timeval_argument(char **tokens, size_t ntoks,
-						struct timeval *tval)
+static int nilfs_cldconfig_get_time_argument(char **tokens, size_t ntoks,
+					     struct timespec *ts)
 {
 	unsigned long num;
 	double fnum;
@@ -156,15 +157,15 @@ static int nilfs_cldconfig_get_timeval_argument(char **tokens, size_t ntoks,
 			goto failed;
 		if (errno == ERANGE)
 			goto failed_too_large;
-		tval->tv_sec = num;
-		tval->tv_usec = (fnum - num) * 1000000;
+		ts->tv_sec = num;
+		ts->tv_nsec = (fnum - num) * 1000000000L;
 	} else if (*endptr != '\0') {
 		goto failed;
 	} else {
 		if (num == ULONG_MAX)
 			goto failed_too_large;
-		tval->tv_sec = num;
-		tval->tv_usec = 0;
+		ts->tv_sec = num;
+		ts->tv_nsec = 0;
 	}
 	return 0;
 
@@ -273,10 +274,59 @@ nilfs_cldconfig_handle_protection_period(struct nilfs_cldconfig *config,
 					 char **tokens, size_t ntoks,
 					 struct nilfs *nilfs)
 {
-	return nilfs_cldconfig_get_timeval_argument(
+	return nilfs_cldconfig_get_time_argument(
 		tokens, ntoks, &config->cf_protection_period);
 }
 
+static unsigned long long
+nilfs_convert_units_to_bytes(const struct nilfs_param *param)
+{
+	unsigned long long bytes = param->num;
+
+	switch (param->unit) {
+	case NILFS_SIZE_UNIT_KB:
+		bytes *= 1000ULL;
+		break;
+	case NILFS_SIZE_UNIT_KIB:
+		bytes <<= 10;
+		break;
+	case NILFS_SIZE_UNIT_MB:
+		bytes *= 1000000ULL;
+		break;
+	case NILFS_SIZE_UNIT_MIB:
+		bytes <<= 20;
+		break;
+	case NILFS_SIZE_UNIT_GB:
+		bytes *= 1000000000ULL;
+		break;
+	case NILFS_SIZE_UNIT_GIB:
+		bytes <<= 30;
+		break;
+	case NILFS_SIZE_UNIT_TB:
+		bytes *= 1000000000000ULL;
+		break;
+	case NILFS_SIZE_UNIT_TIB:
+		bytes <<= 40;
+		break;
+	case NILFS_SIZE_UNIT_PB:
+		bytes *= 1000000000000000ULL;
+		break;
+	case NILFS_SIZE_UNIT_PIB:
+		bytes <<= 50;
+		break;
+	case NILFS_SIZE_UNIT_EB:
+		bytes *= 1000000000000000000ULL;
+		break;
+	case NILFS_SIZE_UNIT_EIB:
+		bytes <<= 60;
+		break;
+	default:
+		assert(0);
+	}
+
+	return bytes;
+}
+
 static unsigned long long
 nilfs_convert_size_to_nsegments(struct nilfs *nilfs, struct nilfs_param *param)
 {
@@ -287,48 +337,7 @@ nilfs_convert_size_to_nsegments(struct nilfs *nilfs, struct nilfs_param *param)
 	} else if (param->unit == NILFS_SIZE_UNIT_PERCENT) {
 		ret = (nilfs_get_nsegments(nilfs) * param->num + 99) / 100;
 	} else {
-		bytes = param->num;
-
-		switch (param->unit) {
-		case NILFS_SIZE_UNIT_KB:
-			bytes *= 1000ULL;
-			break;
-		case NILFS_SIZE_UNIT_KIB:
-			bytes <<= 10;
-			break;
-		case NILFS_SIZE_UNIT_MB:
-			bytes *= 1000000ULL;
-			break;
-		case NILFS_SIZE_UNIT_MIB:
-			bytes <<= 20;
-			break;
-		case NILFS_SIZE_UNIT_GB:
-			bytes *= 1000000000ULL;
-			break;
-		case NILFS_SIZE_UNIT_GIB:
-			bytes <<= 30;
-			break;
-		case NILFS_SIZE_UNIT_TB:
-			bytes *= 1000000000000ULL;
-			break;
-		case NILFS_SIZE_UNIT_TIB:
-			bytes <<= 40;
-			break;
-		case NILFS_SIZE_UNIT_PB:
-			bytes *= 1000000000000000ULL;
-			break;
-		case NILFS_SIZE_UNIT_PIB:
-			bytes <<= 50;
-			break;
-		case NILFS_SIZE_UNIT_EB:
-			bytes *= 1000000000000000000ULL;
-			break;
-		case NILFS_SIZE_UNIT_EIB:
-			bytes <<= 60;
-			break;
-		default:
-			assert(0);
-		}
+		bytes = nilfs_convert_units_to_bytes(param);
 		segment_size = nilfs_get_block_size(nilfs) *
 			nilfs_get_blocks_per_segment(nilfs);
 		ret = (bytes + segment_size - 1) / segment_size;
@@ -367,24 +376,15 @@ nilfs_cldconfig_handle_clean_check_interval(struct nilfs_cldconfig *config,
 					    char **tokens, size_t ntoks,
 					    struct nilfs *nilfs)
 {
-	return nilfs_cldconfig_get_timeval_argument(
+	return nilfs_cldconfig_get_time_argument(
 		tokens, ntoks, &config->cf_clean_check_interval);
 }
 
-static unsigned long long
-nilfs_cldconfig_selection_policy_timestamp(const struct nilfs_suinfo *si)
-{
-	return si->sui_lastmod;
-}
-
 static int
-nilfs_cldconfig_handle_selection_policy_timestamp(struct nilfs_cldconfig *config,
+nilfs_cldconfig_handle_selection_policy_timestamp(struct nilfs_cldconfig *cf,
 						  char **tokens, size_t ntoks)
 {
-	config->cf_selection_policy.p_importance =
-		NILFS_CLDCONFIG_SELECTION_POLICY_IMPORTANCE;
-	config->cf_selection_policy.p_threshold =
-		NILFS_CLDCONFIG_SELECTION_POLICY_THRESHOLD;
+	cf->cf_selection_policy = NILFS_SELECTION_POLICY_TIMESTAMP;
 	return 0;
 }
 
@@ -455,12 +455,59 @@ nilfs_cldconfig_handle_mc_nsegments_per_clean(struct nilfs_cldconfig *config,
 	return 0;
 }
 
+static unsigned long long
+nilfs_convert_size_to_blocks_per_segment(struct nilfs *nilfs,
+					 struct nilfs_param *param)
+{
+	unsigned long long ret, segment_size, block_size, bytes;
+
+	if (param->unit == NILFS_SIZE_UNIT_NONE) {
+		ret = param->num;
+	} else if (param->unit == NILFS_SIZE_UNIT_PERCENT) {
+		ret = (nilfs_get_blocks_per_segment(nilfs) * param->num + 99)
+				/ 100;
+	} else {
+		block_size = nilfs_get_block_size(nilfs);
+		segment_size = block_size *
+				nilfs_get_blocks_per_segment(nilfs);
+		bytes = nilfs_convert_units_to_bytes(param) % segment_size;
+		ret = (bytes + block_size - 1) / block_size;
+	}
+	return ret;
+}
+
+static int
+nilfs_cldconfig_handle_min_reclaimable_blocks(struct nilfs_cldconfig *config,
+					      char **tokens, size_t ntoks,
+					      struct nilfs *nilfs)
+{
+	struct nilfs_param param;
+
+	if (nilfs_cldconfig_get_size_argument(tokens, ntoks, &param) == 0)
+		config->cf_min_reclaimable_blocks =
+			nilfs_convert_size_to_blocks_per_segment(nilfs, &param);
+	return 0;
+}
+
+static int
+nilfs_cldconfig_handle_mc_min_reclaimable_blocks(struct nilfs_cldconfig *config,
+						 char **tokens, size_t ntoks,
+						 struct nilfs *nilfs)
+{
+	struct nilfs_param param;
+
+	if (nilfs_cldconfig_get_size_argument(tokens, ntoks, &param) == 0)
+		config->cf_mc_min_reclaimable_blocks =
+			nilfs_convert_size_to_blocks_per_segment(nilfs, &param);
+	return 0;
+}
+
 static int
 nilfs_cldconfig_handle_cleaning_interval(struct nilfs_cldconfig *config,
 					 char **tokens, size_t ntoks,
 					 struct nilfs *nilfs)
 {
-	return nilfs_cldconfig_get_timeval_argument(
+	return nilfs_cldconfig_get_time_argument(
 		tokens, ntoks, &config->cf_cleaning_interval);
 }
 
@@ -469,7 +516,7 @@ nilfs_cldconfig_handle_mc_cleaning_interval(struct nilfs_cldconfig *config,
 					    char **tokens, size_t ntoks,
 					    struct nilfs *nilfs)
 {
-	return nilfs_cldconfig_get_timeval_argument(
+	return nilfs_cldconfig_get_time_argument(
 		tokens, ntoks, &config->cf_mc_cleaning_interval);
 }
 
@@ -478,7 +525,7 @@ nilfs_cldconfig_handle_retry_interval(struct nilfs_cldconfig *config,
 				      char **tokens, size_t ntoks,
 				      struct nilfs *nilfs)
 {
-	return nilfs_cldconfig_get_timeval_argument(
+	return nilfs_cldconfig_get_time_argument(
 		tokens, ntoks, &config->cf_retry_interval);
 }
 
@@ -490,6 +537,14 @@ static int nilfs_cldconfig_handle_use_mmap(struct nilfs_cldconfig *config,
 	return 0;
 }
 
+static int nilfs_cldconfig_handle_use_set_suinfo(struct nilfs_cldconfig *config,
+						 char **tokens, size_t ntoks,
+						 struct nilfs *nilfs)
+{
+	config->cf_use_set_suinfo = 1;
+	return 0;
+}
+
 static const struct nilfs_cldconfig_log_priority
 nilfs_cldconfig_log_priority_table[] = {
 	{"emerg",	LOG_EMERG},
@@ -502,22 +557,19 @@ nilfs_cldconfig_log_priority_table[] = {
 	{"debug",	LOG_DEBUG},
 };
 
-#define NILFS_CLDCONFIG_NLOGPRIORITIES			\
-	(sizeof(nilfs_cldconfig_log_priority_table) /	\
-	 sizeof(nilfs_cldconfig_log_priority_table[0]))
-
 static int
 nilfs_cldconfig_handle_log_priority(struct nilfs_cldconfig *config,
 				    char **tokens, size_t ntoks,
 				    struct nilfs *nilfs)
 {
+	const struct nilfs_cldconfig_log_priority *clp;
 	int i;
 
-	for (i = 0; i < NILFS_CLDCONFIG_NLOGPRIORITIES; i++) {
-		if (strcmp(tokens[1],
-			   nilfs_cldconfig_log_priority_table[i].cl_name) == 0) {
-			config->cf_log_priority =
-				nilfs_cldconfig_log_priority_table[i].cl_priority;
+	clp = nilfs_cldconfig_log_priority_table;
+	for (i = 0; i < ARRAY_SIZE(nilfs_cldconfig_log_priority_table);
+	     i++, clp++) {
+		if (strcmp(tokens[1], clp->cl_name) == 0) {
+			config->cf_log_priority = clp->cl_priority;
 			return 0;
 		}
 	}
@@ -576,12 +628,20 @@ nilfs_cldconfig_keyword_table[] = {
 		"log_priority", 2, 2,
 		nilfs_cldconfig_handle_log_priority
 	},
+	{
+		"min_reclaimable_blocks", 2, 2,
+		nilfs_cldconfig_handle_min_reclaimable_blocks
+	},
+	{
+		"mc_min_reclaimable_blocks", 2, 2,
+		nilfs_cldconfig_handle_mc_min_reclaimable_blocks
+	},
+	{
+		"use_set_suinfo", 1, 1,
+		nilfs_cldconfig_handle_use_set_suinfo
+	},
 };
 
-#define NILFS_CLDCONFIG_NKEYWORDS			\
-	(sizeof(nilfs_cldconfig_keyword_table) /	\
-	 sizeof(nilfs_cldconfig_keyword_table[0]))
-
 static int nilfs_cldconfig_handle_keyword(struct nilfs_cldconfig *config,
 					  char **tokens, size_t ntoks,
 					  struct nilfs *nilfs)
@@ -593,7 +653,7 @@ static int nilfs_cldconfig_handle_keyword(struct nilfs_cldconfig *config,
 		return 0;
 
 	kw = nilfs_cldconfig_keyword_table;
-	for (i = 0; i < NILFS_CLDCONFIG_NKEYWORDS; i++, kw++) {
+	for (i = 0; i < ARRAY_SIZE(nilfs_cldconfig_keyword_table); i++, kw++) {
 		if (strcmp(tokens[0], kw->ck_text) == 0) {
 			if (check_tokens(tokens, ntoks, kw->ck_minargs,
 					 kw->ck_maxargs) < 0)
@@ -611,12 +671,10 @@ static void nilfs_cldconfig_set_default(struct nilfs_cldconfig *config,
 {
 	struct nilfs_param param;
 
-	config->cf_selection_policy.p_importance =
-		NILFS_CLDCONFIG_SELECTION_POLICY_IMPORTANCE;
-	config->cf_selection_policy.p_threshold =
-		NILFS_CLDCONFIG_SELECTION_POLICY_THRESHOLD;
+	config->cf_selection_policy = NILFS_SELECTION_POLICY_TIMESTAMP;
+
 	config->cf_protection_period.tv_sec = NILFS_CLDCONFIG_PROTECTION_PERIOD;
-	config->cf_protection_period.tv_usec = 0;
+	config->cf_protection_period.tv_nsec = 0;
 
 	param.num = NILFS_CLDCONFIG_MIN_CLEAN_SEGMENTS;
 	param.unit = NILFS_CLDCONFIG_MIN_CLEAN_SEGMENTS_UNIT;
@@ -628,19 +686,32 @@ static void nilfs_cldconfig_set_default(struct nilfs_cldconfig *config,
 	config->cf_max_clean_segments =
 		nilfs_convert_size_to_nsegments(nilfs, &param);
 
-	config->cf_clean_check_interval.tv_sec = NILFS_CLDCONFIG_CLEAN_CHECK_INTERVAL;
-	config->cf_clean_check_interval.tv_usec = 0;
+	config->cf_clean_check_interval.tv_sec =
+		NILFS_CLDCONFIG_CLEAN_CHECK_INTERVAL;
+	config->cf_clean_check_interval.tv_nsec = 0;
 	config->cf_nsegments_per_clean = NILFS_CLDCONFIG_NSEGMENTS_PER_CLEAN;
 	config->cf_mc_nsegments_per_clean =
 		NILFS_CLDCONFIG_MC_NSEGMENTS_PER_CLEAN;
 	config->cf_cleaning_interval.tv_sec = NILFS_CLDCONFIG_CLEANING_INTERVAL;
-	config->cf_cleaning_interval.tv_usec = 0;
-	config->cf_mc_cleaning_interval.tv_sec = NILFS_CLDCONFIG_MC_CLEANING_INTERVAL;
-	config->cf_mc_cleaning_interval.tv_usec = 0;
+	config->cf_cleaning_interval.tv_nsec = 0;
+	config->cf_mc_cleaning_interval.tv_sec =
+		NILFS_CLDCONFIG_MC_CLEANING_INTERVAL;
+	config->cf_mc_cleaning_interval.tv_nsec = 0;
 	config->cf_retry_interval.tv_sec = NILFS_CLDCONFIG_RETRY_INTERVAL;
-	config->cf_retry_interval.tv_usec = 0;
+	config->cf_retry_interval.tv_nsec = 0;
 	config->cf_use_mmap = NILFS_CLDCONFIG_USE_MMAP;
+	config->cf_use_set_suinfo = NILFS_CLDCONFIG_USE_SET_SUINFO;
 	config->cf_log_priority = NILFS_CLDCONFIG_LOG_PRIORITY;
+
+	param.num = NILFS_CLDCONFIG_MIN_RECLAIMABLE_BLOCKS;
+	param.unit = NILFS_CLDCONFIG_MIN_RECLAIMABLE_BLOCKS_UNIT;
+	config->cf_min_reclaimable_blocks =
+		nilfs_convert_size_to_blocks_per_segment(nilfs, &param);
+
+	param.num = NILFS_CLDCONFIG_MC_MIN_RECLAIMABLE_BLOCKS;
+	param.unit = NILFS_CLDCONFIG_MC_MIN_RECLAIMABLE_BLOCKS_UNIT;
+	config->cf_mc_min_reclaimable_blocks =
+		nilfs_convert_size_to_blocks_per_segment(nilfs, &param);
 }
 
 static inline int iseol(int c)
@@ -682,19 +753,23 @@ static int nilfs_cldconfig_do_read(struct nilfs_cldconfig *config,
 	char line[LINE_MAX], *tokens[NTOKENS_MAX];
 	int ntoks, c, ret;
 
-	if ((fp = fopen(conffile, "r")) == NULL)
+	fp = fopen(conffile, "r");
+	if (!fp)
 		return -1;
 
 	ret = 0;
 	while (fgets(line, sizeof(line), fp) != NULL) {
 		if (line[strlen(line) - 1] != '\n') {
 			syslog(LOG_WARNING, "too long line");
-			while (((c = fgetc(fp)) != '\n') && (c != EOF)) ;
+			while (((c = fgetc(fp)) != '\n') && (c != EOF))
+				;
 		}
-		if ((ntoks = tokenize(line, tokens, NTOKENS_MAX)) == 0)
+		ntoks = tokenize(line, tokens, NTOKENS_MAX);
+		if (ntoks == 0)
 			continue;
-		if ((ret = nilfs_cldconfig_handle_keyword(
-			     config, tokens, ntoks, nilfs)) < 0)
+		ret = nilfs_cldconfig_handle_keyword(config, tokens, ntoks,
+						     nilfs);
+		if (ret < 0)
 			break;
 	}
 
